#include <eZ8.h>
#include <stdlib.h>
#include "GrLCD.h"
#include "Font8.h"
#include "Font16.h"

/***************************************
	lcd display constants
***************************************/
#define GR_MODE		0		// mode register
#define GR_PITCH	1		// character H & V pitch register
#define GR_HORIZ	2		// horiz chars or bytes register
#define GR_DUTY		3		// duty cycle register
#define GR_CURP		4		// cursor posn register
#define GR_STRTL	8		// display start addr L register
#define GR_STRTH	9		// display start addr H register
#define GR_CURAL	0xA		// cursor addr L register
#define GR_CURAH	0xB		// cursor addr H register
#define GR_WRITE	0xC		// write instruction
#define GR_READ		0xD		// read instruction
#define GR_BCLR		0xE		// bit clear instruction
#define GR_BSET		0xF		// bit set instruction

static const BYTE JustTbl[8] = {
		0, NR_COLS/4,					// L00J,L25J
		NR_COLS/3, NR_COLS/2,			// L33J,C50J
		(NR_COLS*2)/3, (NR_COLS*3)/4,	// R66J,R75J
		NR_COLS, 0						// R100J,ABSJ
};

BYTE CJ_code = 0;			// control/justify code for text display
BYTE linenr;				// starting line for text
BYTE startcol;				// starting col for later use
BYTE endcol;				// ending column
int gr_x = 0;
int gr_y = 0;

static const WORD mask[8] = { 0xFF, 0x7F, 0x3F, 0x1F, 0xF, 7, 3, 1 };

static const BYTE LCDIni[10][2] = {
		GR_MODE, 0x12,
		GR_PITCH, 0x77,
		GR_HORIZ, 19,
		GR_DUTY, 159,
		GR_CURP, 0,
		GR_STRTL, 0,
		GR_STRTH, 0,
		GR_CURAL, 0,
		GR_CURAH, 0,
		GR_MODE, 0x32
};

static void delay(int dly)
{
	while(dly > 0)
	{
		--dly;
	}
}

/***********************************************************
	Initialize display
***********************************************************/

void LCDInit(void)
{
	BYTE a;

	PCOUT = 0;				// assert hardware reset
	delay(1000);
	PCOUT = RES;			// de-assert reset

	for (a = 0; a < 10; a++)
	{
		LCDSet(LCDIni[a][0], LCDIni[a][1]);	
	}
	LCDClear(0, NR_ROWS);			// clear data mamory
}

/***********************************************************
	Check LCD controller busy status.
***********************************************************/

static void LCDBusy(void)
{
	PEDD = 0xFF;				//set bus port to inputs

	PCOUT = RW + RS + RES;		// set for status read
	DELAY;
	PCOUT = E + RW + RS + RES;	// assert enable
	DELAY;

	while (LCD_IN & 0x80);		// wait for busy bit to go low

	PCOUT = RW + RS + RES;		// de-assert enable
	PCOUT = RES;				// set for write

	PEDD = 0;					//set bus port to outputs
}

/***********************************************************
	Write command + data bytes to LCD controller
***********************************************************/

void LCDSet(BYTE cmd, BYTE arg)
{
	LCDBusy();					// wait til not busy, if necessary
	PCOUT = RS + RES;			// select command entry
	LCD_OUT = cmd;				// put byte on the bus
	DELAY;
	PCOUT = E + RS + RES;		// assert enable
	DELAY;
	PCOUT = RS + RES;			// de-assert enable
	DELAY;

	PCOUT = RES;				// select data entry
	LCD_OUT = arg;				// put byte on the bus
	DELAY;
	PCOUT = E + RES;			// assert enable
	DELAY;
	PCOUT = RES;				// de-assert enable
}

/***********************************************************
	Write command byte to LCD controller
***********************************************************/

void LCDCmd(BYTE arg)
{
	LCDBusy();					// wait til not busy, if necessary
	PCOUT = RS + RES;			// select command entry
	LCD_OUT = arg;				// put byte on the bus
	DELAY;
	PCOUT = E + RS + RES;		// assert enable
	DELAY;
	PCOUT = RS + RES;			// de-assert enable
}

/***********************************************************
	Write data byte to LCD controller
***********************************************************/

void LCDData(BYTE arg)
{
	LCDBusy();					// wait til not busy, if necessary
	PCOUT = RES;				// select data entry
	LCD_OUT = arg; 				// put byte on the bus
	DELAY;
	PCOUT = E + RES;			// assert enable for write
	DELAY;
	PCOUT = RES;				// de-assert enable
}

/***********************************************************
	Read byte from LCD controller
***********************************************************/

BYTE LCDRead(void)
{
	BYTE arg;
	LCDBusy();					// wait til not busy, if necessary
	LCDCmd(GR_READ);

	PEDD = 0xFF;	//here set your LCD data bus port to inputs
	PCOUT = RW + RES;		// select data
	PCOUT = E + RW + RES;	// assert enable
	DELAY;
	arg = LCD_IN;			// get byte from the bus
	PCOUT = RW + RES;		// de-assert enable
	PCOUT = RES;			// write
	PEDD = 0;		//here set your LCD data bus port to outputs
	return (arg);
}

/***********************************************************
	Clear the display data RAM
	srow = starting row
	hgt = number of rows
***********************************************************/

void LCDClear(BYTE srow, BYTE hgt)
{
	int a, b;

	LCDSetCur((WORD)srow * NR_BYTES);
	LCDCmd(GR_WRITE);
	PCOUT = RES;
	b = (int)NR_BYTES * (int)hgt;
	LCD_OUT = 0;				// put byte on the bus

	for (a = 0; a < b; a++)
	{
		DELAY;
		PCOUT = E + RES;		// assert enable
		DELAY;
		PCOUT = RES;			// de-assert enable
		LCDBusy();
	}
}

/***********************************************************
	Write text string to display.
	Sample: String_T str1 = { C50J+FONT16, 2, 0, "Hello World" };

	strg = pointer to String_T structure
***********************************************************/

void LCDText(String_T *strg)
{
	BYTE a, b, c, txtlen, chrlen, fbyte, lbyte;
	BYTE hght, chr, row, col, offs;
	BYTE buff[20];
	WORD cur_addr, chrbits, temp;
	const WORD rom *bits;
	const char rom *cp;

	CJ_code = strg->cj_code;		// save CJ code
	linenr = strg->row;				// save starting row number
	b = CJ_code & JUSTIFY;			// isolate justify code
	hght = CJ_code & FONT16 ? 16 : 8;

	txtlen = LCDPixLen(strg->text);// get display length in pixels

	if (b == L00J)
	{
		col = 0;
	}
	else if (b == ABSJ)
	{
		col = strg->col;
	}
	else
	{
		if (txtlen > NR_COLS)
		{
			txtlen = NR_COLS;
		}
		a = txtlen;
		if (b != R100J)				// if not right justify
		{
			a /= 2;					// use half length for centering
		}
		col = JustTbl[b] - a;		// set starting column
	}

	if (col >= NR_COLS ||
		col + txtlen > NR_COLS)		// if over/underflow
	{
		col = 0;					// use left justify
		if (txtlen > NR_COLS - 1)
		{
			txtlen = NR_COLS - 1;
		}
	}
	startcol = col;					// save start column
	endcol = col + txtlen;			// save end column
	fbyte = col / 8;				// first byte on display line
	lbyte = endcol / 8;				// last byte on display line

	for (row = 0; row < hght; row++)	// for each row of the text
	{
		col = startcol;					// starting col
		offs = col % 8;					// starting offset
		cur_addr = (WORD)((WORD)(linenr + row) * NR_BYTES) + fbyte;
		LCDSetCur(cur_addr);
		LCDRead();						// dummy read
	
		for (a = fbyte, b = 0; a <= lbyte; a++)
		{
			buff[b++] = LCDRead();		// read row data from display
		}
	
		cp = strg->text;				// get pointer to the text
		b = 0;							// index into buff[]
	
		while (*cp)
		{
			chr = *cp++;				// get next text character
			if (CJ_code & FONT16)		// large font
			{
				if (chr < firstchr_L)	// char code out of range
				{
					chr = firstchr_L;
				}
				a = chr - firstchr_L;
				if (a >= nr_chrs_L)		// char code out of range
				{
					a = 0;
				}

				chrlen = (rom *)lentbl_L[a];	// number of columns
				bits = (rom *)chrtbl_L[a];		// bitmap data pointer
			}
			else						// small font
			{
				if (chr < firstchr_S)	// char code out of range
				{
					chr = firstchr_S;
				}
				a = chr - firstchr_S;
				if (a >= nr_chrs_S)		// char code out of range
				{
					a = 0;
				}
			
				chrlen = (rom *)lentbl_S[a];	// number of columns
				bits = (rom *)chrtbl_S[a];		// bitmap data pointer
			}
			chrbits = bits[row];		// get row data for char
			c = chrlen;					// temporary char length
			if (! offs)					// aligned with byte boundary
			{
				buff[b] |= chrbits & 0xFF;
				if (c > 8)				// do remaining byte
				{
					buff[++b] |= chrbits >> 8;
				}
			}
			else						// need to shift char bits
			{
				a = 8 - offs;			// # bits to use
				temp = chrbits & mask[offs];
				temp <<= offs;			// shift masked bits
				buff[b] |= temp;
				if (c > a)				// more bits left?
				{
					c -= a;
					chrbits >>= a;		// shift out used bits
					buff[++b] |= chrbits & 0xFF;
					if (c >= 8)			// do any remaining bits
					{
						chrbits >>= 8;
						buff[++b] |= chrbits & 0xFF;
					}
				}
			}
			col += chrlen;
			if (! (CJ_code & NOSPC))
			{
				col++;					// add blank column for spacing
			}
			b = (col / 8) - fbyte;		// be sure buff index is correct
			offs = col % 8;				// bit offset to next column
		}	// while (*cp)

		LCDSetCur(cur_addr);			// set addr to write data
		LCDCmd(GR_WRITE);				// send write instruction
	
		for (a = fbyte, b = 0; a <= lbyte; a++)
		{
			LCDData(buff[b++]);			// send new row data to display
		}
	}	// for (row

}

/***********************************************************
	Erase a block of pixels
***********************************************************/

void LCDErase(BYTE srow, BYTE scol, BYTE erow, BYTE ecol)
{
	int a, b, c;

	for (c = (int)srow; c <= (int)erow; c++)
	{
		LCDSetCur(c * NR_BYTES + (scol / 8));
		LCDCmd(GR_WRITE);
		PCOUT = RES;
		b = (int)((ecol / 8) - (scol / 8)) + 1;
		LCD_OUT = 0;				// put byte on the bus
	
		for (a = 0; a < b; a++)
		{
			DELAY;
			PCOUT = E + RES;		// assert enable
			DELAY;
			PCOUT = RES;			// de-assert enable
			LCDBusy();
		}
	}
}

/***********************************************************
	Scroll a text string
***********************************************************/

void LCDScrollText(BYTE srow, BYTE scol, BYTE erow, BYTE ecol, String_T *txt, int rate)
{
	BYTE xrow, xcol;
	int time;

	xrow = srow;
	xcol = scol;
	txt->cj_code = (txt->cj_code & ~JUSTIFY) | ABSJ;	// insure absolute col mode

	for (;;)
	{
		txt->row = xrow;
		txt->col = xcol;
		LCDText(txt);
		if (xrow == erow && xcol == ecol) break;
		for (time = 0; time < rate; time++)
		{
			delay(1000);
		}
		LCDErase(xrow, startcol, xrow + (CJ_code & FONT16 ? 16 : 8), endcol);
		if (xrow != erow)
		{
			if (xrow < erow) xrow++;
			else xrow--;
		}
		if (xcol != ecol)
		{
			if (xcol < ecol) xcol++;
			else xcol--;
		}
	}
}

/***********************************************************
	Write cursor address to controller
***********************************************************/

void LCDSetCur(WORD addr)
{
	LCDSet(GR_CURAL, addr & 0xFF);
	LCDSet(GR_CURAH, addr >> 8);
}

/***********************************************************
	Get length of display text, in pixels
***********************************************************/

BYTE LCDPixLen(const char *chr)
{
	BYTE len, first, nrc, txt;
	const BYTE *lentbl;

	if (CJ_code & FONT16)		// large font
	{
		first = firstchr_L;
		lentbl = (const BYTE *)lentbl_L;
		nrc = nr_chrs_L;
	}
	else						// small font
	{
		first = firstchr_S;
		lentbl = (const BYTE *)lentbl_S;
		nrc = nr_chrs_S;
	}

	len = 0;

	while (*chr)
	{
		txt = *chr;
		if (txt >= first && (txt - first) < nrc)
		{
			len += lentbl[txt - first];

			if (! (CJ_code & NOSPC))
			{
				len++;					// include blank between chars
			}
		}
		chr++;
	}
	return len;
}


/***********************************************************
	Plot a pixel at X,Y non-destructively
	0,0 is at upper right
***********************************************************/

void plot(int x, int y)
{
	WORD cur_addr;
	BYTE data;

	if (x >= NR_COLS || y >= NR_ROWS) return;

	cur_addr = (y * NR_BYTES) + (x / 8);
	LCDSetCur(cur_addr);	// address the display byte
	LCDRead();				// dummy read
	data = LCDRead();		// read display data
	data |= 1 << (x % 8);	// OR in new bit
	LCDSetCur(cur_addr);	// address the display byte
	LCDSet(GR_WRITE, data);	// write it back to display
}

/***********************************************************
	Plot a pixel at X,Y and save point
***********************************************************/

void point(int x, int y)
{
	gr_x = x;
	gr_y = y;
	plot(x, y);
}

/***********************************************************
	Draw a line from saved X,Y to X2,Y2 and save new endpoint
	Bresenham's algorithm
***********************************************************/

void lineto(int x2, int y2)
{
	int i, delta_x, delta_y, numpixels;
	int d, dinc1, dinc2;
	int x, xinc1, xinc2;
	int y, yinc1, yinc2;

	// Calculate delta_x and delta_y for initialisation
	delta_x = abs(x2 - gr_x);
	delta_y = abs(y2 - gr_y);

	// Initialize all vars based on which is the independent variable
	if (delta_x >= delta_y)		// x is independent variable
	{
		numpixels = delta_x + 1;
		d = (delta_y << 1) - delta_x;
		dinc1 = delta_y << 1;
		dinc2 = (delta_y - delta_x) << 1;
		xinc1 = 1;
		xinc2 = 1;
		yinc1 = 0;
		yinc2 = 1;
	}
	else			// y is independent variable
	{
		numpixels = delta_y + 1;
		d = (delta_x << 1) - delta_y;
		dinc1 = delta_x << 1;
		dinc2 = (delta_x - delta_y) << 1;
		xinc1 = 0;
		xinc2 = 1;
		yinc1 = 1;
		yinc2 = 1;
	}


	if (gr_x > x2)	// Make sure x and y move in the right directions
	{
		xinc1 = - xinc1;
		xinc2 = - xinc2;
	}
	if (gr_y > y2)
	{
		yinc1 = - yinc1;
		yinc2 = - yinc2;
	}


	x = gr_x;		// Start drawing from previous endpoint
	y = gr_y;
	gr_x = x2;		// save new endpoint for next line
	gr_y = y2;

	// Draw the pixels
	for (i = 0; i < numpixels; i++)
	{
		plot(x, y);
		if (d < 0)
		{
			d = d + dinc1;
			x = x + xinc1;
			y = y + yinc1;
		}
		else
		{
			d = d + dinc2;
			x = x + xinc2;
			y = y + yinc2;
		}
	}
}
